use 5.014;
use strict;
use warnings;

package MooX::Pression;

use Zydeco ();

our $AUTHORITY = 'cpan:TOBYINK';
our $VERSION   = '0.508';
our @ISA       = 'Zydeco';

*PACKAGE_SPEC =\&Zydeco::PACKAGE_SPEC;

1;

__END__

=pod

=encoding utf-8

=head1 NAME

MooX::Pression - old name of Zydeco

=head1 DESCRIPTION

This package exists for backwards compatibility.

=head1 SEE ALSO

L<Zydeco>.

=head1 AUTHOR

Toby Inkster E<lt>tobyink@cpan.orgE<gt>.

=head1 COPYRIGHT AND LICENCE

This software is copyright (c) 2020 by Toby Inkster.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=head1 DISCLAIMER OF WARRANTIES

THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

